import sys
import os
import platform
import shutil

from silly_db.helpers import set_executable


help = """
PLOP
Provide a Local Operationnal Pack

'python3 -m silly_db.plop [option]'

options are:


[db]
provides::
- a 'database' directory containing:
    - a 'migrations' directory containing 2 examples: inital.sql, populate.sql
- a 'main_example.py' that could be a starting point for your project.
"""

BASE_DIR = os.path.abspath(os.path.dirname(__file__))
cwd = os.getcwd()
this_os = platform.system()


def get_db():
    """Copy a starter kit in the working directory in a more convenient
    way for work"""
    shutil.copy(BASE_DIR+'/db/main_example.py', cwd)
    shutil.copytree(
        BASE_DIR+'/db/database', os.path.join(cwd, "database"))
    print("files successfully acquired")
    # if not running windows, set executable:
    if not this_os.lower().startswith("w"):
        set_executable(os.path.join(cwd, 'main_example.py'))


options = {
    'db': get_db,
}


if __name__ == '__main__':
    if len(sys.argv) == 1:
        print(help)

    if len(sys.argv) == 2:
        arg = sys.argv[1].lower()
        if arg not in options:
            print("invalid option given to silly_db.plop")
            print(help)
        else:
            options[arg]()



#! /usr/bin/env python3
# coding: utf-8

"""Example of an embryonic main.py file, this should be included in
a MCV designed program"""

import os
from silly_db.db import DB

BASE_DIR = os.path.abspath(os.path.dirname(__file__))

db = DB(
    base=BASE_DIR,
    file="database/my_db.sqlite3",
    migrations_dir="database/migrations")
db.migrate_all()



import os
import stat
import hashlib


def hasher(file):
    """Returns the sha1 hash of a file"""
    BLOCK_SIZE = 65536
    hash = hashlib.sha1()
    with open(file, 'r') as f:
        fb = f.read(BLOCK_SIZE).encode('utf-8')
        while len(fb) > 0:
            hash.update(fb)
            fb = f.read(BLOCK_SIZE).encode('utf-8')
    return hash.hexdigest()


def set_executable(file) -> None:
    """set a file executable, used within plop"""
    st = os.stat(file)
    os.chmod(file, st.st_mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)


def insert_to_sql(value):
    """Cleans the input data to be insertable into sql.
    - str: Fixes the quote problem.
    - None: becomes NULL
    - other: keep it as it is.
    """

    if isinstance(value, type(None)):
        sql_value = 'NULL'
    elif type(value) == str:
        sql_value = value.replace(
                    "'", "''")
        sql_value = f"'{sql_value}'"
    else:
        sql_value = value
    return sql_value


# color parameters: style;background (30 is none);foreground
color = {
    "end": "\x1b[0m",
    "info": "\x1b[0;30;36m",
    "success": "\x1b[0;30;32m",
    "warning": "\x1b[0;30;33m",
    "danger": "\x1b[0;30;31m",
}


INITIALIZE_DB = """
CREATE TABLE IF NOT EXISTS "_migrations_applied" (
    "id" INTEGER NOT NULL,
    "file" TEXT NOT NULL,
    "date" TIMESTAMP DEFAULT CURRENT_TIMESTAMP NOT NULL,
    "sha1" TEXT NOT NULL,
    PRIMARY KEY("id" AUTOINCREMENT)
);

"""



import os
import platform
import shutil

from silly_db.helpers import set_executable
from silly_db import __version__

from flamewok.cli import cli


BASE_DIR = os.path.abspath(os.path.dirname(__file__))
cwd = os.getcwd()
this_os = platform.system()


plop_help = """
PLOP
Provide a Local Operationnal Pack

To get a starter pack in the current working directory:

$ silly-db plop db

provides:
- a 'database' directory containing:
    - a 'migrations' directory containing 2 examples: inital.sql, populate.sql
- a 'main_example.py' that could be a starting point for your project.
"""


def display_plop_help(*args):
    print(plop_help)


def get_db(*args):
    """Copy a starter kit in the working directory in a more convenient
    way for work"""
    shutil.copy(BASE_DIR+'/plop/db/main_example.py', cwd)
    shutil.copytree(
        BASE_DIR+'/plop/db/database', os.path.join(cwd, "database"))
    print("files successfully acquired")
    # if not running windows, set executable:
    if not this_os.lower().startswith("w"):
        set_executable(os.path.join(cwd, 'main_example.py'))


def cmd():
    cli.route(
        "HELP :",
        ("", cli.help, "display this help"),
        ("-h", cli.help, "idem"),
        ("--help", cli.help, "idem"),
        "PLOP :",
        ("plop", display_plop_help, "About plop"),
        ("plop db", get_db,
            "get a starter pack in the current working directory"),
        "_"*60,
        "ABOUT:",
        "package: silly-db",
        f"version: {__version__}",
        "description: Very light ORM for SQLite, simple and efficient",
        "home page: https://github.com/byoso/silly_db/",
    )


if __name__ == "__main__":
    cmd()



"""Objects and tools used for the selections from the database.
- Selection
- SelectionItem
- Model
"""

from silly_db.exceptions import SillyDbError
from silly_db.helpers import color, insert_to_sql


class Selection:
    """Object returned by the DB.select method, it contains the result
    of the query in a convenient form allowing easy manipulations.
    - self.items is a list of SelectionItems."""
    def __init__(self, *args):
        self.items = [*args]

    def __str__(self) -> str:
        display = "<Selection["
        for elem in self.items:
            display += str(elem)+", "
        display += "]>"
        return display

    def __len__(self):
        return len(self.items)

    def __getitem__(self, key):
        return self.items[key]

    def __iter__(self):
        return iter(self.items)

    def __add__(self, other):
        if isinstance(other, SelectionItem):
            self.items.append(other)
            return self
        array = [*self, *other]
        return Selection(*array)

    def jsonify(self) -> list:
        """Returns an array of SelectionItems turned into dicts"""
        array = []
        for item in self.items:
            array.append(dict(item))
        return array

    def exists(self) -> bool:
        if len(self.items) > 0:
            return True
        return False

    def order_by(self, key=None, reverse=False, case=False):
        if case:
            return Selection(*sorted(
                self.items, key=lambda x: getattr(x, key), reverse=reverse))
        else:
            return Selection(*sorted(
                self.items, key=lambda x: getattr(
                    x, key).upper(), reverse=reverse))


class SelectionItem:
    """db.select() will build SelectionItems with some given attributes
    and create a Selection object, filled with SelectionItems"""
    def __init__(self, dico=None, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)
        if dico is not None:
            for key, value in dico.items():
                setattr(self, key, value)

    def __str__(self) -> str:
        return f"<SelItem{str(dict(self))}>"

    def __repr__(self) -> str:
        return str(self)

    def __iter__(self):
        """dict(SelectionItem) converts the
        SelectionItem into a dict recursively"""
        for attr in vars(self):
            object = getattr(self, attr)
            if isinstance(object, SelectionItem):
                object = dict(object)
            if isinstance(object, Selection):
                object = object.jsonify()
            yield attr, object

    def __add__(self, other):
        selection = Selection() + self + other
        return selection

    def _copy(self):
        new = SelectionItem()
        for attr in vars(self):
            setattr(new, attr, getattr(self, attr))
        return new

    def remove(self, *args):
        """remove one or more attributes"""
        for arg in args:
            delattr(self, arg)
        return self

    def jsonify(self) -> dict:
        return dict(self)


class Model:
    """DB.model() instanciates a Model
    - db is provided automaticaly by db.model(), do not care
    - table is a str (the name of your SQL table)
    """
    def __init__(self, db, table):
        self._db = db
        self._table = table

    def all(self, columns="*"):
        "returns a Selection of all the rows in the table"
        selection = self._db.select(f"{columns} FROM {self._table}")
        return selection

    def filter(self, condition, columns="*"):
        """returns a Selection"""
        selection = self._db.select(
            f"{columns} FROM {self._table} WHERE {condition}"
            )
        return selection

    def get(self, condition, columns="*"):
        """returns a unique SelectionItem object, or an exception if
        multiple results, or None if no result"""
        selection = self.filter(condition, columns)
        if len(selection) > 1:
            raise SillyDbError(
                "Model.get returning multiple items. The condition "
                "must target only one result, you should use a unique id."
                )
        if not selection.exists():
            return None
        else:
            return selection[0]

    def get_id(self, id, columns="*"):
        """Shortcut to self.get("id=...")"""
        if id is None:
            return None
        return self.get(f"id={id}", columns)

    def insert(self, **kwargs):
        if len(kwargs) == 0 and self._db.debug:
            message = (
                color['warning'] + "Warning : Empty insert in model" +
                color['end']
                )
            print(message)
        keys = ""
        values = ""
        for key in kwargs:
            keys += key + ", "
            value = insert_to_sql(kwargs[key])
            values += f"{value}, "
        keys = keys[:-2]
        values = values[:-2]
        command = (
            f"INSERT OR REPLACE INTO {self._table} ({keys}) VALUES ({values})"
            )

        self._db.cursor.execute("BEGIN TRANSACTION;")
        self._db.cursor.execute(command)
        self._db.cursor.execute("COMMIT;")

    def delete(self, condition=None):
        if condition is not None:
            command = f"DELETE FROM {self._table} WHERE {condition}"
            self._db.cursor.execute("BEGIN TRANSACTION;")
            self._db.cursor.execute(command)
            self._db.cursor.execute("COMMIT;")

    def update(self, condition=None, **kwargs):
        command = f"UPDATE {self._table} SET "
        values = ""
        for key in kwargs:
            values += f"{key}={insert_to_sql(kwargs[key])}, "
        values = values[:-2]
        command += values + " WHERE " + condition
        self._db.cursor.execute("BEGIN TRANSACTION;")
        self._db.cursor.execute(command)
        self._db.cursor.execute("COMMIT;")




"""This module provides the class DB, main class of silly_db.

"""


import os
import sqlite3
from silly_db.helpers import (
    color,
    INITIALIZE_DB,
    hasher,
    )
from silly_db.exceptions import (
    SillyDbError,
)

from silly_db.selections import (
    Selection,
    SelectionItem,
    Model,
)


class DB:
    """Central object interacting with the database itself.
    params:
        - base: the absolute path on wich 'file' and 'migrations_dir' are
        relatives to.
        - file (required): path to your desired [name].sqlite3 file
        - migrations_dir: path to the directory containing the migrations
        - debug: default is 'True', shows some warnings in console.
         'False' will hide the warnings.
    """
    def __init__(
        self,
        base=None,
        file: str = None,
        migrations_dir: str = None,
        debug: bool = True
            ):
        if file is None:
            raise SillyDbError(
                "Missing parameter for DB: file=some_file_name.sqlite3")
        self.file = os.path.join(base, file)
        self.connection = sqlite3.connect(self.file)
        self.cursor = self.connection.cursor()
        if migrations_dir is None:
            self.migrations_dir = None
        else:
            self.migrations_dir = os.path.join(base, migrations_dir)
        self.debug = debug

        self.execute(INITIALIZE_DB)

    def _read_sql_file(self, file: str) -> str:
        """Returns the content of a given file"""
        with open(file, 'r') as sql_file:
            commands = sql_file.read()
        return commands

    def _get_headers(self, cursor_description):
        """Gathers the headers from a query. In case of duplicated headers,
        renames them, and sends a warning (if self.debug is True)"""
        headers = [description[0] for description in cursor_description]
        uniques = []
        warning = False  # inform the user if the headers have been changed
        for header in headers:
            counter = 1
            if header not in uniques:
                uniques.append(header)
            else:
                uniques.append(header + f"_{counter}")
                counter += 1
                warning = True
        if warning and self.debug:
            print(
                color["warning"] +
                "\n!!! WARNING : ambiguous query :\n"
                "Headers with same names, silly_db arbitrarily "
                "changed some of your headers. Use '[name] as [other_name)]'"
                " in your SQL query to fix this.\n" +
                color['end']
                )
        return uniques

    def model(self, table):
        """Takes a table name as parameter, returns the table as a Model"""
        new_model = Model(self, table)
        return new_model

    def execute(self, command):
        """'Begin transaction' and 'commit' are automatically added
        to the command, so don't use this key words."""
        commands = command.strip().split(";")
        # execution
        self.cursor.execute("BEGIN TRANSACTION;")
        for command in commands:
            self.cursor.execute(command)
        self.cursor.execute("COMMIT;")

    def migrate(self, file=None):
        """Execute the migrations from a .sql file if the migration
        have not be done already."""
        hashes_selection = self.select("sha1 FROM _migrations_applied")
        hashes = [hash.sha1 for hash in hashes_selection]
        sha1 = hasher(file)
        if sha1 not in hashes:
            commands = self._read_sql_file(file)

            self.connection.executescript(commands)
            file_name = os.path.split(file)[-1]
            register_migration = (
                "INSERT INTO '_migrations_applied' (file, sha1) "
                f"VALUES('{file_name}', '{sha1}');"
            )
            self.execute(register_migration)
            if self.debug:
                print(
                    color["success"] +
                    f"Migration successfully applied: {file_name}" +
                    color['end']
                    )

    def migrate_all(self, directory=None):
        """migrates all the files in the self.migrations_dir or a given
        directory"""
        if directory is not None:
            apply_dir = directory
        elif self.migrations_dir is not None:
            apply_dir = self.migrations_dir
        else:
            print("Unable to migrate: No migration directory has been given")
        files = [
            file for file in os.listdir(apply_dir)
            if file.endswith(".sql")
            ]
        files.sort()
        for file in files:
            self.migrate(os.path.abspath(os.path.join(apply_dir, file)))

    def export(
        self,
        structure_file="00_structure.sql_bkp",
        data_file="01_data.sql_bkp",
        action="both"  # both / structure / data
            ):
        """By default exports both data and structure
        with default file names"""
        structure = ""
        data = "BEGIN TRANSACTION;\n"
        for line in self.connection.iterdump():
            if "sqlite_sequence" not in line and \
                    "_migrations_applied" not in line:
                if line.startswith("INSERT"):
                    data += line+"\n"
                else:
                    structure += line+"\n"
        data += "COMMIT;"
        if action == "both" or action == "structure":
            with open(structure_file, 'w') as f:
                f.write(structure)
        if action == "both" or action == "data":
            with open(data_file, 'w') as f:
                f.write(data)

    def export_all(self, file=None):
        """Exports both structure and data to a single file,
         into a sql format."""
        if file is None:
            file = "backup_all_in_one.sql_bkp"
        with open(file, 'w') as f:
            for line in self.connection.iterdump():
                if "sqlite_sequence" not in line and \
                        "_migrations_applied" not in line:
                    f.write(f"{line}\n")

    def export_data(self, file=None):
        """Exports the datas only to a given file,
         into a sql format."""
        if file is not None:
            self.export(data_file=file, action="data")
        else:
            self.export(action="data")

    def export_structure(self, file=None):
        """Exports only the structure of the database to a given file,
        into a sql format."""
        if file is None:
            file = "00_structure.sql_bkp"
        self.cursor.execute("SELECT sql FROM sqlite_master;")
        with open(file, 'w') as f:
            f.write("BEGIN TRANSACTION;\n")
            for line in self.cursor.fetchall():
                if "sqlite_sequence" not in line[0] and \
                        "_migrations_applied" not in line[0]:
                    f.write(f"{line[0]};\n")
            f.write("COMMIT;")

    def query(self, command: str) -> Selection:
        """command is a SQL query, returns a Selection object"""
        self.cursor.execute(command)
        headers = self._get_headers(self.cursor.description)
        results = self.cursor.fetchall()
        answers = []
        responses = []
        for result in results:
            elem = dict(zip(headers, result))
            answers.append(elem)
        for answer in answers:
            dico = dict(answer.items())
            response = SelectionItem(dico)
            responses.append(response)
        selection = Selection(*responses)
        return selection

    def select(self, command: str) -> Selection:
        """Query tool, the parameter is a purely SQL query, but without
        the word SELECT at the beginig (as it's already the method name)
        """
        command = "SELECT " + command
        return self.query(command)
